/*******************************************************************************
 mod_h264_streaming.c

 mod_h264_streaming - An Apache module for pseudo-streaming Quicktime/MPEG4 files.
 http://h264.code-shop.com

 Copyright (C) 2008-2009 CodeShop B.V.
******************************************************************************/ 

#include <httpd.h>
#include <http_core.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_log.h>
#include <apr_buckets.h>
#include "moov.h"

// Add the following two lines to your httpd config.
// LoadModule h264_streaming_module /usr/lib/apache2/modules/mod_h264_streaming.so
// AddHandler h264-streaming.extensions .mp4

#define H264_STREAMING_HANDLER "h264-streaming.extensions"

static char* get_param_value(request_rec *r, char* name)
{
  if(r->args)
  {
    char* p = strstr(r->args, name);
    if(p)
    {
      p = p + strlen(name);
      return p;
    }
  }

  return NULL;
}

static double get_param(request_rec *r, char* name)
{
  double value = 0.0;

  char* p = get_param_value(r, name);
  if(p)
  {
    value = strtod(p, NULL);
  }

  return value;
}

static int drive_h264_streaming(request_rec *r)
{
  apr_finfo_t fi;
  apr_bucket_brigade *bb;
  apr_file_t *fp = NULL;
  apr_status_t rv = APR_SUCCESS;
  
  rv = apr_stat(&fi, r->filename, APR_FINFO_SIZE, r->pool);
  
  if (rv)
  {
    /* Let the core handle it. */
    return DECLINED;
  }
  
  /* Open the file */
  rv = apr_file_open(&fp, r->filename, APR_READ, APR_OS_DEFAULT, r->pool);
  
  if (rv)
  {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
                  "file permissions deny server access: %s", r->filename);
    return HTTP_FORBIDDEN;
  }

  {
    unsigned int filesize = fi.size;
    double start = get_param(r, "start=");
    double end = get_param(r, "end=");
    char* client_value = get_param_value(r, "client=");
    int client_is_flash = client_value != NULL ? starts_with(client_value, "FLASH") : 0;

    void* mp4_header;
    uint32_t mp4_header_size;
    uint64_t mdat_offset;
    uint64_t mdat_size;

    int result = mp4_split(r->filename, filesize, start, end,
                           &mp4_header, &mp4_header_size,
                           &mdat_offset, &mdat_size, client_is_flash);

    if(result)
    {
      bb = apr_brigade_create(r->pool, r->connection->bucket_alloc);
      rv = apr_brigade_write(bb, NULL, NULL, mp4_header, mp4_header_size);
      if (rv)
      {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
                      "unable to write mp4 header in brigade");
        return HTTP_INTERNAL_SERVER_ERROR;
      }

      apr_brigade_insert_file(bb, fp, mdat_offset, mdat_size, r->pool);
    }

    if(mp4_header)
    {
      free(mp4_header);
    }

    if(!result)
    {
      return DECLINED;
    }

    ap_set_content_type(r, "video/mp4");
    ap_set_content_length(r, mp4_header_size + mdat_size);
    if(!client_is_flash)
    {
      apr_table_set(r->headers_out, "X-Mod-H264-Streaming", "version=2.0");
    }
    else
    {
      apr_table_set(r->headers_out, "X-Mod-H264-Streaming", "version=2.0,client=flash");
    }
  }
  
  return ap_pass_brigade(r->output_filters, bb);
}

static int h264_streaming_handler(request_rec *r)
{
  if ((!r->handler) || (strcmp(r->handler, H264_STREAMING_HANDLER)))
  {
    return DECLINED;
  }
  
  r->allowed |= (AP_METHOD_BIT << M_GET);
  if (r->method_number != M_GET)
  {
    return HTTP_METHOD_NOT_ALLOWED;
  }
  
  return drive_h264_streaming(r);
}

static const command_rec h264_streaming_cmds[] =
{
    {NULL}
};

static void register_hooks(apr_pool_t *p)
{
  ap_hook_handler(h264_streaming_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA h264_streaming_module =
{
  STANDARD20_MODULE_STUFF,
  NULL,
  NULL,
  NULL,
  NULL,
  h264_streaming_cmds,
  register_hooks
};

// End Of File

